#include "stdafx.h"
#include "ButtonTexture.h"

NameValuePair modeConstants[] =
	{
		{ "stretch", MULTIBLT_STRETCH },
		{ "tile", MULTIBLT_TILE },
		{ "tile-horizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tile-vertical", MULTIBLT_TILEVERTICAL },
		{ "tilehorizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tilevertical", MULTIBLT_TILEVERTICAL },
		{ 0, 0 }
		};

ButtonTexture::ButtonTexture(const string &name)
{
	memoryNormalDC = 0;
	memoryNormalBitmap = 0;

	memoryActiveDC = 0;
	memoryActiveBitmap = 0;

	memoryMinimizedDC = 0;
	memoryMinimizedBitmap = 0;

	memoryFlashingDC = 0;
	memoryFlashingBitmap = 0; 

	memoryHoverDC = 0;
	memoryHoverBitmap = 0; 

	memoryGroupDC = 0;
	memoryGroupBitmap = 0; 

	/*has_normalAlphaMap = false;
	has_activeAlphaMap = false;
	has_minimizedAlphaMap = false;
	has_flashingAlphaMap = false;*/

	defaultbutton.button_saturation = GetxRCInt(name , "ButtonSaturationIntensity", 255, 0, 255);
	defaultbutton.button_hueIntensity = GetxRCInt(name , "ButtonHueIntensity", 0, 0, 255);
	defaultbutton.button_hueColor = GetxRCColor(name, "ButtonHueColor", RGB(255, 255, 255));

	defaultbutton.Alpha = GetxRCInt(name , "ButtonAlphaBlend", 255, 0, 255);

	defaultbutton.leftEdge = GetxRCInt(name, "ButtonImageLeftEdge", 0, 0);
	defaultbutton.topEdge = GetxRCInt(name, "ButtonImageTopEdge", 0, 0);
	defaultbutton.rightEdge = GetxRCInt(name, "ButtonImageRightEdge", 0, 0);
	defaultbutton.bottomEdge = GetxRCInt(name, "ButtonImageBottomEdge", 0, 0);

	defaultbutton.mode = GetxRCNamedValue(name, "ButtonImageMode", modeConstants, MULTIBLT_STRETCH);

	defaultsolidcolors = GetxRCLine(name, "ButtonSolidColors", "ffffff");
	defaultbevelSize = GetxRCInt(name, "ButtonSolidBevelSize", 0, 0);
	defaultgradientColor = GetxRCLine(name, "ButtonSolidGradientColors", "");
	defaultverticalGradient = GetxRCBool(name, "ButtonSolidGradientVertical");
}

ButtonTexture::~ButtonTexture()
{
	NormalGradient.clear();
	ActiveGradient.clear();
	MinimizedGradient.clear();
	FlashingGradient.clear();

	HoverGradient.clear();
	GroupGradient.clear();

	if(memoryNormalBitmap)
	{
		memoryNormalBitmap = (HBITMAP) SelectObject(memoryNormalDC, memoryNormalBitmap);
		DeleteObject(memoryNormalBitmap);
	}
	if(memoryNormalDC)
		DeleteDC(memoryNormalDC);

	if(memoryActiveBitmap)
	{
		memoryActiveBitmap = (HBITMAP) SelectObject(memoryActiveDC, memoryActiveBitmap);
		DeleteObject(memoryActiveBitmap);
	}
	if(memoryActiveDC)
		DeleteDC(memoryActiveDC);

	if(memoryMinimizedBitmap)
	{
		memoryMinimizedBitmap = (HBITMAP) SelectObject(memoryMinimizedDC, memoryMinimizedBitmap);
		DeleteObject(memoryMinimizedBitmap);
	}
	if(memoryMinimizedDC)
		DeleteDC(memoryMinimizedDC);

	if(memoryFlashingBitmap)
	{
		memoryFlashingBitmap = (HBITMAP) SelectObject(memoryFlashingDC, memoryFlashingBitmap);
		DeleteObject(memoryFlashingBitmap);
	}
	if(memoryFlashingDC)
		DeleteDC(memoryFlashingDC);

	if(memoryHoverBitmap)
	{
		memoryHoverBitmap = (HBITMAP) SelectObject(memoryHoverDC, memoryHoverBitmap);
		DeleteObject(memoryHoverBitmap);
	}
	if(memoryHoverDC)
		DeleteDC(memoryHoverDC);

	if(memoryGroupBitmap)
	{
		memoryGroupBitmap = (HBITMAP) SelectObject(memoryGroupDC, memoryGroupBitmap);
		DeleteObject(memoryGroupBitmap);
	}
	if(memoryGroupDC)
		DeleteDC(memoryGroupDC);
}

void ButtonTexture::parseCFG(string cfg, COLORREF &bgcolor, COLORREF &lightcolor, COLORREF &darkcolor)
{
	int color;
	char main[32], light[32], dark[32];
	char *buffers[] = {main, light, dark};

	int numTokens = LCTokenize(cfg.c_str(), buffers, 3, NULL);

	bgcolor = RGB(255,255,255);
	lightcolor = RGB(255,255,255);
	darkcolor = RGB(255,255,255);
		
	if(numTokens == 1)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	else if (numTokens == 2)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	else if (numTokens == 3)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(dark, NULL, 16);
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
}

void ButtonTexture::parseGradientCFG(string cfg, int imageType)
{
	int color;
	COLORREF insertColor;
	char token[4096] = {0};
	LPCSTR nextToken = cfg.c_str();
	
	while (GetToken(nextToken, token, &nextToken, false))
	{
		color = strtol(token, NULL, 16);
		insertColor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		if (imageType == 0)
		{
			NormalGradient.insert(NormalGradient.end(), insertColor );
		}
		else if (imageType == 1)
		{
			ActiveGradient.insert(ActiveGradient.end(), insertColor );
		}
		else if (imageType == 2)
		{
			MinimizedGradient.insert(MinimizedGradient.end(), insertColor );
		}
		else if (imageType == 3)
		{
			FlashingGradient.insert(FlashingGradient.end(), insertColor );
		}
		else if (imageType == 4)
		{
			HoverGradient.insert(HoverGradient.end(), insertColor );
		}
		else if (imageType == 5)
		{
			GroupGradient.insert(GroupGradient.end(), insertColor );
		}
	}
}

void ButtonTexture::subconfigure(string type, BUTTONSETTINGS &workSettings, BUTTONSETTINGS &defaultSettings)
{
	workSettings.button_saturation = GetxRCInt(type , "ButtonSaturationIntensity", defaultSettings.button_saturation, 0, 255);
	workSettings.button_hueIntensity = GetxRCInt(type , "ButtonHueIntensity", defaultSettings.button_hueIntensity, 0, 255);
	workSettings.button_hueColor = GetxRCColor(type, "ButtonHueColor", defaultSettings.button_hueColor);

	workSettings.Alpha = GetxRCInt(type , "ButtonAlphaBlend", defaultSettings.Alpha, 0, 255);

	workSettings.leftEdge = GetxRCInt(type, "ButtonImageLeftEdge", defaultSettings.leftEdge, 0);
	workSettings.topEdge = GetxRCInt(type, "ButtonImageTopEdge", defaultSettings.topEdge, 0);
	workSettings.rightEdge = GetxRCInt(type, "ButtonImageRightEdge", defaultSettings.rightEdge, 0);
	workSettings.bottomEdge = GetxRCInt(type, "ButtonImageBottomEdge", defaultSettings.bottomEdge, 0);

	workSettings.mode = GetxRCNamedValue(type, "ButtonImageMode", modeConstants, defaultSettings.mode);
}

void ButtonTexture::configure(const string &name)
{
	string imageFile = "";

	usenormalSolidColors = GetxRCBool(name, "UseNormalButtonSolidColors", GetxRCBool(name, "UseButtonSolidColors"));

	normalsolidcolors = defaultsolidcolors;

	if (usenormalSolidColors)
	{
		normalsolidcolors = GetxRCLine(name, "NormalButtonSolidColors", defaultsolidcolors);
		parseCFG(normalsolidcolors, normalbackgroundColor, normallightBevelColor, normaldarkBevelColor);
		//BG Insert in Gradient List
		NormalGradient.insert(NormalGradient.end(), normalbackgroundColor );

		normalbevelSize = GetxRCInt(name, "NormalButtonSolidBevelSize", defaultbevelSize, 0);

		normalgradientColor = GetxRCLine(name, "NormalButtonSolidGradientColors", defaultgradientColor);
		parseGradientCFG(normalgradientColor, 0);
		
		normalverticalGradient = GetxRCBool(name, "NormalButtonSolidGradientVertical", defaultverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "ButtonImage", "");
		if (imageFile.empty())
			imageFile = GetxRCString(name, "NormalButtonImage", "");
		else if (!imageFile.empty())
		{
			string temp = GetxRCString(name, "NormalButtonImage", "");
			if ( !temp.empty() )
				imageFile = temp;
		}

		memoryNormalBitmap = LoadLSImage(imageFile.c_str(), 0);

		/*if (memoryNormalBitmap)
		{
			BITMAP bm;
			GetObject(memoryNormalBitmap, sizeof(BITMAP), &bm);
			has_normalAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

			has_activeAlphaMap = has_minimizedAlphaMap = has_flashingAlphaMap = has_normalAlphaMap;

			if (has_normalAlphaMap)
				PrepareAlphaChannel(memoryNormalBitmap);
		}*/

		GetLSBitmapSize(memoryNormalBitmap, &imageNormalWidth, &imageNormalHeight);

		memoryNormalDC = CreateCompatibleDC(0);
		SelectObject(memoryNormalDC, memoryNormalBitmap);
	}

	useactiveSolidColors = GetxRCBool(name, "UseActiveButtonSolidColors", usenormalSolidColors);

	if (useactiveSolidColors)
	{
		imageFile = GetxRCLine(name, "ActiveButtonSolidColors", normalsolidcolors);
		parseCFG(imageFile, activebackgroundColor, activelightBevelColor, activedarkBevelColor);
		//BG Insert in Gradient List
		ActiveGradient.insert(ActiveGradient.end(), activebackgroundColor );

		activebevelSize = GetxRCInt(name, "ActiveButtonSolidBevelSize", normalbevelSize, 0);

		parseGradientCFG(GetxRCLine(name, "ActiveButtonSolidGradientColors", normalgradientColor), 1);
		
		activeverticalGradient = GetxRCBool(name, "ActiveButtonSolidGradientVertical", normalverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "ActiveButtonImage", "");
		if (!imageFile.empty())
		{
			memoryActiveBitmap = LoadLSImage(imageFile.c_str(), 0);

			/*if (memoryActiveBitmap)
			{
				BITMAP bm;
				GetObject(memoryActiveBitmap, sizeof(BITMAP), &bm);
				has_activeAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

				if (has_activeAlphaMap)
					PrepareAlphaChannel(memoryActiveBitmap);
			}*/

			GetLSBitmapSize(memoryActiveBitmap, &imageActiveWidth, &imageActiveHeight);

			memoryActiveDC = CreateCompatibleDC(0);
			SelectObject(memoryActiveDC, memoryActiveBitmap);
		}
	}

	useminimizedSolidColors = GetxRCBool(name, "UseMinimizedButtonSolidColors", usenormalSolidColors);

	if (useminimizedSolidColors)
	{
		imageFile = GetxRCLine(name, "MinimizedButtonSolidColors", normalsolidcolors);
		parseCFG(imageFile, minimizedbackgroundColor, minimizedlightBevelColor, minimizeddarkBevelColor);
		//BG Insert in Gradient List
		MinimizedGradient.insert(MinimizedGradient.end(), minimizedbackgroundColor );

		minimizedbevelSize = GetxRCInt(name, "MinimizedButtonSolidBevelSize", normalbevelSize, 0);

		parseGradientCFG(GetxRCLine(name, "MinimizedButtonSolidGradientColors", normalgradientColor), 2);

		minimizedverticalGradient = GetxRCBool(name, "MinimizedButtonSolidGradientVertical", normalverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "MinimizedButtonImage", "");
		if (!imageFile.empty())
		{
			memoryMinimizedBitmap = LoadLSImage(imageFile.c_str(), 0);

			/*if (memoryMinimizedBitmap)
			{
				BITMAP bm;
				GetObject(memoryMinimizedBitmap, sizeof(BITMAP), &bm);
				has_minimizedAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

				if (has_minimizedAlphaMap)
					PrepareAlphaChannel(memoryMinimizedBitmap);
			}*/

			GetLSBitmapSize(memoryMinimizedBitmap, &imageMinimizedWidth, &imageMinimizedHeight);

			memoryMinimizedDC = CreateCompatibleDC(0);
			SelectObject(memoryMinimizedDC, memoryMinimizedBitmap);
		}
	}

	useflashingSolidColors = GetxRCBool(name, "UseFlashingButtonSolidColors", usenormalSolidColors);

	if (useflashingSolidColors)
	{
		imageFile = GetxRCLine(name, "FlashingButtonSolidColors", normalsolidcolors);
		parseCFG(imageFile, flashingbackgroundColor, flashinglightBevelColor, flashingdarkBevelColor);
		//BG Insert in Gradient List
		FlashingGradient.insert(FlashingGradient.end(), flashingbackgroundColor );

		flashingbevelSize = GetxRCInt(name, "FlashingButtonSolidBevelSize", normalbevelSize, 0);
	
		parseGradientCFG(GetxRCLine(name, "FlashingButtonSolidGradientColors", normalgradientColor), 3);

		flashingverticalGradient = GetxRCBool(name, "FlashingButtonSolidGradientVertical", normalverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "FlashingButtonImage", "");
		if (!imageFile.empty())
		{
			memoryFlashingBitmap = LoadLSImage(imageFile.c_str(), 0);

			/*if (memoryFlashingBitmap)
			{
				BITMAP bm;
				GetObject(memoryFlashingBitmap, sizeof(BITMAP), &bm);
				has_flashingAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

				if (has_flashingAlphaMap)
					PrepareAlphaChannel(memoryFlashingBitmap);
			}*/

			GetLSBitmapSize(memoryFlashingBitmap, &imageFlashingWidth, &imageFlashingHeight);

			memoryFlashingDC = CreateCompatibleDC(0);
			SelectObject(memoryFlashingDC, memoryFlashingBitmap);
		}
	}

	usehoverSolidColors = GetxRCBool(name, "UseHoverButtonSolidColors");

	if (usehoverSolidColors)
	{
		imageFile = GetxRCLine(name, "HoverButtonSolidColors", normalsolidcolors);
		parseCFG(imageFile, hoverbackgroundColor, hoverlightBevelColor, hoverdarkBevelColor);
		//BG Insert in Gradient List
		HoverGradient.insert(HoverGradient.end(), hoverbackgroundColor );

		hoverbevelSize = GetxRCInt(name, "HoverButtonSolidBevelSize", normalbevelSize, 0);

		parseGradientCFG(GetxRCLine(name, "HoverButtonSolidGradientColors", normalgradientColor), 4);

		hoververticalGradient = GetxRCBool(name, "HoverButtonSolidGradientVertical", normalverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "HoverButtonImage", "");
		if (!imageFile.empty())
		{
			memoryHoverBitmap = LoadLSImage(imageFile.c_str(), 0);

			/*if (memoryHoverBitmap)
			{
				BITMAP bm;
				GetObject(memoryHoverBitmap, sizeof(BITMAP), &bm);
				has_hoverAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

				if (has_hoverAlphaMap)
					PrepareAlphaChannel(memoryHoverBitmap);
			}*/

			GetLSBitmapSize(memoryHoverBitmap, &imageHoverWidth, &imageHoverHeight);

			memoryHoverDC = CreateCompatibleDC(0);
			SelectObject(memoryHoverDC, memoryHoverBitmap);
		}
	}

	usegroupSolidColors = GetxRCBool(name, "UseGroupedButtonSolidColors");

	if (usegroupSolidColors)
	{
		imageFile = GetxRCLine(name, "GroupedButtonSolidColors", normalsolidcolors);
		parseCFG(imageFile, groupbackgroundColor, grouplightBevelColor, groupdarkBevelColor);
		//BG Insert in Gradient List
		GroupGradient.insert(GroupGradient.end(), groupbackgroundColor );

		groupbevelSize = GetxRCInt(name, "GroupedButtonSolidBevelSize", normalbevelSize, 0);

		parseGradientCFG(GetxRCLine(name, "GroupedButtonSolidGradientColors", normalgradientColor), 5);

		groupverticalGradient = GetxRCBool(name, "GroupedButtonSolidGradientVertical", normalverticalGradient);
	}
	else
	{
		imageFile = GetxRCString(name, "GroupedButtonImage", "");
		if (!imageFile.empty())
		{
			memoryGroupBitmap = LoadLSImage(imageFile.c_str(), 0);

			/*if (memoryGroupBitmap)
			{
				BITMAP bm;
				GetObject(memoryGroupBitmap, sizeof(BITMAP), &bm);
				has_groupAlphaMap = (bm.bmBits != 0) && (bm.bmBitsPixel == 32);

				if (has_groupAlphaMap)
					PrepareAlphaChannel(memoryGroupBitmap);
			}*/

			GetLSBitmapSize(memoryGroupBitmap, &imageGroupWidth, &imageGroupHeight);

			memoryGroupDC = CreateCompatibleDC(0);
			SelectObject(memoryGroupDC, memoryGroupBitmap);
		}
	}

	subconfigure((name + "Normal"), normal, defaultbutton);
	normalAlpha = normal.Alpha;

	subconfigure((name + "Active"), active, normal);
	activeAlpha = active.Alpha;

	subconfigure((name + "Minimized"), minimized, normal);
	minimizedAlpha = minimized.Alpha;

	subconfigure((name + "Flashing"), flashing, normal);
	flashingAlpha = flashing.Alpha;

	subconfigure((name + "Hover"), hover, normal);
	hoverAlpha = hover.Alpha;

	subconfigure((name + "Grouped"), group, normal);
	groupAlpha = group.Alpha;

	if ((normal.button_saturation != 255 || normal.button_hueIntensity != 0) /*&& !has_normalAlphaMap*/ && !usenormalSolidColors)
		paintEffects(memoryNormalDC, imageNormalWidth, imageNormalHeight, normal.button_saturation, normal.button_hueIntensity, normal.button_hueColor);
	if (memoryActiveDC && (active.button_saturation != 255 || active.button_hueIntensity != 0) /*&& !has_activeAlphaMap*/ && !useactiveSolidColors)
		paintEffects(memoryActiveDC, imageActiveWidth, imageActiveHeight, active.button_saturation, active.button_hueIntensity, active.button_hueColor);
	if (memoryMinimizedDC && (minimized.button_saturation != 255 || minimized.button_hueIntensity != 0) /*&& !has_minimizedAlphaMap*/ && !useminimizedSolidColors)
		paintEffects(memoryMinimizedDC, imageMinimizedWidth, imageMinimizedHeight, minimized.button_saturation, minimized.button_hueIntensity, minimized.button_hueColor);
	if (memoryFlashingDC && (flashing.button_saturation != 255 || flashing.button_hueIntensity != 0) /*&& !has_flashingAlphaMap*/ && !useflashingSolidColors)
		paintEffects(memoryFlashingDC, imageFlashingWidth, imageFlashingHeight, flashing.button_saturation, flashing.button_hueIntensity, flashing.button_hueColor);
	if (memoryHoverDC && (hover.button_saturation != 255 || hover.button_hueIntensity != 0) /*&& !has_hoverAlphaMap*/ && !usehoverSolidColors)
		paintEffects(memoryHoverDC, imageHoverWidth, imageHoverHeight, hover.button_saturation, hover.button_hueIntensity, hover.button_hueColor);
	if (memoryGroupDC && (group.button_saturation != 255 || group.button_hueIntensity != 0) /*&& !has_groupAlphaMap*/ && !usegroupSolidColors)
		paintEffects(memoryGroupDC, imageGroupWidth, imageGroupHeight, group.button_saturation, group.button_hueIntensity, group.button_hueColor);
}

void ButtonTexture::apply(HDC hDC, int x, int y, int width, int height, int buttonType)
{
	RECT r;
	r.left = x;
	r.top = y;
	r.right = r.left + width;
	r.bottom = r.top + height;

	if (buttonType == 0 && usenormalSolidColors)
	{
		// fill in the background
		if (normalbevelSize > 0 && (normalbevelSize*2 < width && normalbevelSize*2 < height))
		{
			::SetBkColor(hDC, normaldarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-normalbevelSize;
			r.bottom = r.bottom-normalbevelSize;
			::SetBkColor(hDC, normallightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+normalbevelSize;
			r.top = r.top+normalbevelSize;

			if ( NormalGradient.size() > 1 )
				solidGradientPaint(hDC, NormalGradient, r, normalverticalGradient);
			else
			{
				::SetBkColor(hDC, normalbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( NormalGradient.size() > 1 )
				solidGradientPaint(hDC, NormalGradient, r, normalverticalGradient);
			else
			{
				::SetBkColor(hDC, normalbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if ( buttonType == 1 && useactiveSolidColors )
	{
		// fill in the background
		if (activebevelSize > 0 && (activebevelSize*2 < width && activebevelSize*2 < height))
		{
			::SetBkColor(hDC, activedarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-activebevelSize;
			r.bottom = r.bottom-activebevelSize;
			::SetBkColor(hDC, activelightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+activebevelSize;
			r.top = r.top+activebevelSize;

			if ( ActiveGradient.size() > 1 )
				solidGradientPaint(hDC, ActiveGradient, r, activeverticalGradient);
			else
			{
				::SetBkColor(hDC, activebackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( ActiveGradient.size() > 1 )
				solidGradientPaint(hDC, ActiveGradient, r, activeverticalGradient);
			else
			{
				::SetBkColor(hDC, activebackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if ( buttonType == 2 && useminimizedSolidColors )
	{
		// fill in the background
		if (minimizedbevelSize > 0 && (minimizedbevelSize*2 < width && minimizedbevelSize*2 < height))
		{
			::SetBkColor(hDC, minimizeddarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-minimizedbevelSize;
			r.bottom = r.bottom-minimizedbevelSize;
			::SetBkColor(hDC, minimizedlightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+minimizedbevelSize;
			r.top = r.top+minimizedbevelSize;

			if ( MinimizedGradient.size() > 1 )
				solidGradientPaint(hDC, MinimizedGradient, r, minimizedverticalGradient);
			else
			{
				::SetBkColor(hDC, minimizedbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( MinimizedGradient.size() > 1 )
				solidGradientPaint(hDC, MinimizedGradient, r, minimizedverticalGradient);
			else
			{
				::SetBkColor(hDC, minimizedbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if ( buttonType == 3 && useflashingSolidColors )
	{
		// fill in the background
		if (flashingbevelSize > 0 && (flashingbevelSize*2 < width && flashingbevelSize*2 < height))
		{
			::SetBkColor(hDC, flashingdarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-flashingbevelSize;
			r.bottom = r.bottom-flashingbevelSize;
			::SetBkColor(hDC, flashinglightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+flashingbevelSize;
			r.top = r.top+flashingbevelSize;

			if ( FlashingGradient.size() > 1 )
				solidGradientPaint(hDC, FlashingGradient, r, flashingverticalGradient);
			else
			{
				::SetBkColor(hDC, flashingbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( FlashingGradient.size() > 1 )
				solidGradientPaint(hDC, FlashingGradient, r, flashingverticalGradient);
			else
			{
				::SetBkColor(hDC, flashingbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if ( buttonType == 4 && usehoverSolidColors )
	{
		// fill in the background
		if (hoverbevelSize > 0 && (hoverbevelSize*2 < width && hoverbevelSize*2 < height))
		{
			::SetBkColor(hDC, hoverdarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-hoverbevelSize;
			r.bottom = r.bottom-hoverbevelSize;
			::SetBkColor(hDC, hoverlightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+hoverbevelSize;
			r.top = r.top+hoverbevelSize;

			if ( HoverGradient.size() > 1 )
				solidGradientPaint(hDC, HoverGradient, r, hoververticalGradient);
			else
			{
				::SetBkColor(hDC, hoverbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( HoverGradient.size() > 1 )
				solidGradientPaint(hDC, HoverGradient, r, hoververticalGradient);
			else
			{
				::SetBkColor(hDC, hoverbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else if ( buttonType == 5 && usegroupSolidColors )
	{
		// fill in the background
		if (groupbevelSize > 0 && (groupbevelSize*2 < width && groupbevelSize*2 < height))
		{
			::SetBkColor(hDC, groupdarkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-groupbevelSize;
			r.bottom = r.bottom-groupbevelSize;
			::SetBkColor(hDC, grouplightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+groupbevelSize;
			r.top = r.top+groupbevelSize;

			if ( GroupGradient.size() > 1 )
				solidGradientPaint(hDC, GroupGradient, r, groupverticalGradient);
			else
			{
				::SetBkColor(hDC, groupbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
		else
		{
			if ( GroupGradient.size() > 1 )
				solidGradientPaint(hDC, GroupGradient, r, groupverticalGradient);
			else
			{
				::SetBkColor(hDC, groupbackgroundColor);
				::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
			}
		}
	}
	else
	{
		if (buttonType == 4 && memoryHoverBitmap == 0)
			return;
		else if (buttonType == 5 && memoryGroupBitmap == 0)
			return;
	
		if ( buttonType == 1 && memoryActiveBitmap == 0 )
			buttonType = 0;
		else if ( buttonType == 2 && memoryMinimizedBitmap == 0 )
			buttonType = 0;
		else if ( buttonType == 3 && memoryFlashingBitmap == 0 )
			buttonType = 0;
		else if ( buttonType == 3 && memoryFlashingBitmap == 0 )
			buttonType = 0;

		if (buttonType == 0 && memoryNormalBitmap == 0)
			return;
		
		HDC bufferDC = CreateCompatibleDC(hDC);
		HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		SelectObject(bufferDC, bufferBitmap);

		if (buttonType == 0)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryNormalDC, 0, 0, imageNormalWidth, imageNormalHeight,
					normal.leftEdge, normal.topEdge, normal.rightEdge, normal.bottomEdge,
					normal.mode, SRCCOPY);
		}
		else if (buttonType == 1)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryActiveDC, 0, 0, imageActiveWidth, imageActiveHeight,
					active.leftEdge, active.topEdge, active.rightEdge, active.bottomEdge,
					active.mode, SRCCOPY);
		}
		else if (buttonType == 2)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryMinimizedDC, 0, 0, imageMinimizedWidth, imageMinimizedHeight,
					minimized.leftEdge, minimized.topEdge, minimized.rightEdge, minimized.bottomEdge,
					minimized.mode, SRCCOPY);
		}
		else if (buttonType == 3)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryFlashingDC, 0, 0, imageFlashingWidth, imageFlashingHeight,
					flashing.leftEdge, flashing.topEdge, flashing.rightEdge, flashing.bottomEdge,
					flashing.mode, SRCCOPY);
		}
		else if (buttonType == 4)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryHoverDC, 0, 0, imageHoverWidth, imageHoverHeight,
					hover.leftEdge, hover.topEdge, hover.rightEdge, hover.bottomEdge,
					hover.mode, SRCCOPY);
		}
		else if (buttonType == 5)
		{
			MultiBlt(bufferDC, 0, 0, width, height,
					memoryGroupDC, 0, 0, imageGroupWidth, imageGroupHeight,
					group.leftEdge, group.topEdge, group.rightEdge, group.bottomEdge,
					group.mode, SRCCOPY);
		}
			
		TransparentBltLS(hDC, x, y, width, height,
					bufferDC, 0, 0,
					RGB(255, 0, 255));

		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);
	}
}

void ButtonTexture::paintEffects(HDC destHDC, int width, int height, int button_saturation, int button_hueIntensity, COLORREF button_hueColor)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	int iy = 0;
	int ix = 0;
	COLORREF cl = 0;
	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	for(iy = 0; iy < height; iy++)
	{
		for(ix = 0; ix < width; ix++)
		{
			cl = GetPixel(destHDC, ix, iy);
			r = GetRValue(cl);
			g = GetGValue(cl);
			b = GetBValue(cl);
		
			if( !( r == 255 && g == 0 && b == 255 ) )
			{
				// Saturation effect
				if(button_saturation != 255)
				{
					BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

					r = (BYTE)((r * button_saturation + gray * (255 - button_saturation) + 255) >> 8);
					g = (BYTE)((g * button_saturation + gray * (255 - button_saturation) + 255) >> 8);
					b = (BYTE)((b * button_saturation + gray * (255 - button_saturation) + 255) >> 8);
				}

				// Hue effect
				if(button_hueIntensity != 0)
				{
					// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
					// test normal
					r = (BYTE)((r * (255 - button_hueIntensity) + GetRValue(button_hueColor) * button_hueIntensity + 255) >> 8);
					g = (BYTE)((g * (255 - button_hueIntensity) + GetGValue(button_hueColor) * button_hueIntensity + 255) >> 8);
					b = (BYTE)((b * (255 - button_hueIntensity) + GetBValue(button_hueColor) * button_hueIntensity + 255) >> 8);
				}
			}

			SetPixelV(destHDC, ix, iy, RGB(r, g, b));
		}
	}
	//End Saturation / Hueing
}

void ButtonTexture::solidGradientPaint(HDC destHDC, vector<COLORREF> workGradient, RECT r, bool vertical)
{
	int nWidth = r.right - r.left;
	int nHeight = r.bottom - r.top;

	RECT r2;
	double steps = 1;
	double cr, cg, cb, dr, dg, db;
	int gradSteps;

	int colorCount = workGradient.size();

	if (vertical)
		gradSteps = max(nHeight / (colorCount-1), 1);
	else
		gradSteps = max(nWidth / (colorCount-1), 1);

	int lastGradientPos;
	if (vertical)
		lastGradientPos = r.top;
	else
		lastGradientPos = r.left;

	for (ColorRefIterator cit = workGradient.begin(); cit != workGradient.end(); cit++ )
	{
		if (cit == workGradient.end())
			break;

		COLORREF one = *cit;
		cit++;
		COLORREF two = *cit;
		cit--;
	
		// get current color
		cr = GetRValue( one );
		cg = GetGValue( one );
		cb = GetBValue( one );

		// Get delta color
		dr = (cr - GetRValue( two )) / gradSteps;
		dg = (cg - GetGValue( two )) / gradSteps;
		db = (cb - GetBValue( two )) / gradSteps;

		for (int i = 1; i <= gradSteps; i++)
		{
			cr -= dr;
			cg -= dg;
			cb -= db;
		
			if (vertical)
				SetRect(&r2, r.left, (int)(lastGradientPos+steps*(i - 1)), r.right, min((int)(lastGradientPos+steps*i), r.bottom));
			else
				SetRect(&r2, (int)(lastGradientPos+steps*(i - 1)), r.top, min((int)(lastGradientPos+steps*i), r.right), r.bottom);

			if (r2.right-r2.left > 0 && r2.bottom-r2.top > 0)
			{
				::SetBkColor(destHDC, RGB((int)cr, (int)cg, (int)cb));
				::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
			}
		}
		if (vertical)
			lastGradientPos = r2.bottom;
		else
			lastGradientPos = r2.right;
	}
}

/*void ButtonTexture::PrepareAlphaChannel(HBITMAP &bmp)
{
	BITMAP bm;
	GetObject(bmp, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	for (int i = 0; i < bm.bmHeight * bm.bmWidth; ++i)
	{
		//Magic Pink
		if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
			pPixel[3] = 0;

		//PreMultiply
		pPixel[0] = pPixel[0] * pPixel[3] / 255;
		pPixel[1] = pPixel[1] * pPixel[3] / 255;
		pPixel[2] = pPixel[2] * pPixel[3] / 255;

		pPixel+= 4;
	}
}*/